' RRBUTTON.INC, CUSTOM CONTROL BUTTON INCLUDE FILE
' Public Domain by Borje Hagsten, September 2001
'
' UPDATES:
'  FEB  3 2003 - Added forwarding of %WM_COMMAND and %WM_NOTIFY to parent,
'                so the control can act as container for other controls too.
'  DEC 30 2002 - Added button return from %WM_GETDLGCODE so it
'                responds properly to mnemonic keyboard input (&)
'                Removed handling of Esc and arrow keys from %WM_KEYDOWN.
'                Fixed a small bug with rr.dn under %RBM_SETLABELMODE
'                It now uses GetWindowLong(hWnd, %GWL_HINSTANCE) to
'                get proper instance handle at creation.
'  DEC 27 2002 - changed some LONG's for handles to DWORD - (PB/WIN 7.0)
'  DEC 13 2002 - added rr.dn reset to %RBM_SETLABELMODE message.
'  FEB 16 2002 - made necessary changes so it compiles with
'                PBDLL 6.1 too. Some pointer errors fixed.
'  SEPTEMBER 12: Discovered wParam in %WM_GETDLGCODE receives
'  keycode, so by trapping this, one can do selective work.
'  Saved some on the keyboard handling by only act on Enter key.
'  SEPTEMBER 11: Added %WM_GETDLGCODE message and now handle all
'  keyboard input self in WM_KEYDOWN.
'  SEPTEMBER 11: Removed SetWindowRegion, only do select into DC's.
'  Also had to change some of the hit testing in mouse events.
'  SEPTEMBER 11: added custom CursorPtInRegion function to handle
'  region rect differences between Win2000 and Win95/98 better.
'  Also added HAND cursor to samples resource file, since older
'  systems may not have this. Changed
'  SEPTEMBER 10: Added %RBM_SETIMGPOSLOCK to enable image to show
'  up/down state by itself. Changed some of the code that handles
'  up/down states, etc, so labels are treated properly in all situations.
'  Added resource file and some smiley icons to show how new message
'  can work with image on a button (here on round button).
'  SEPTEMBER  9: Replaced some of the SetCapture calls with a timer,
'  after tip from Semen Matusovski about problems in Win2K.
'  Also added code in LBUTTONDOWN, to avoid setting focus to labels
'  that are set to show no focus.
'
' INFO:
' Good place for this file is in PB\WINAPI folder, then simply add
' #INCLUDE "RRBUTTON.INC" to enable use of the control. See provided
' sample program for "How to use" information. Look at the messages
' below for explanation on how to set it up.

' Wrote this one to make some things easier in Windows programming.
' Idea is that everything is set up at creation, then button itself
' can handle almost everything while the program is running.
' Same control can be used as Flat button, Raised Button or even as
' inactive/active Label - the later good for things like URL controls, etc.
' The control can position bitmaps, icons and text in exact location, plus
' show different images and text color depending on flat, up or down state.
'
' HINTS:
' you can create very nice buttons by using 3 different images, one
' for flat state (%RBM_SETIMAGE), another for up state (%RBM_SETIMAGEUP)
' and a third for down state (%RBM_SETIMAGEDN). Using RBM_SETCORNER, you
' can adjust the button's region to become anything from square to circle.
' Using %RBM_SETLABELMODE you can create button or label, the later acting
' on all events, but do not become raised/sunken when mouse move over it.
' Using proper images, you can even create nice-looking checkbox with it
' by trapping the %RBN_BTNDN/UP messages and set new image according to
' toggled value, accessible via %RBM_GETTOGGLE, or own flag, whatever.
' You can even use three different text colors, depending on up/down/flat
' state. In short, your own creativity is the limit.. :)
'
' BTW, control has a small palette procedure, to enable somewhat consistent
' look in 256 color mode and the control responds/changes color directly
' on changes in color scheme. No expert in this area, but it seems to work.
' Problem is, individual controls are not told when system palette changes,
' so if another app uses own palette (Netscape is worst), things may look
' a bit bad in low res mode. Not sure how to solve this, but one way seems
' to be by trapping %WM_PALETTECHANGED in main dialog and redraw entire dialog.
' It does the job pretty well and not many uses 256 color mode today anyway, so..
'
' WHISH TO THANK:
' Semen Matusovski, Lance Edmonds, Jules Marchildon and James C. Fuller
' for valuable information on palettes, etc. found in posting to PB's forums.
'
' Not much help here. See sample prog, read comments in code and play
' around with it youself. One thing - have not shown how to load an
' image from file or resource. Following code is all it takes, the rest
' is handled by the control itself, including cleaning up memory:
'-----------------------------------------------------
' bmpFile = "C:\Path\FileName"  '<- can be name in resource file too..
' SendMessage hButn, %RBM_SETIMAGE, %IMAGE_BITMAP, BYVAL STRPTR(bmpFile)
'-----------------------------------------------------

'********************************************************************
' RRBUTTON Custom Control messages and declares
'********************************************************************
%RBM_SETCORNER     = %WM_USER + 1   'wParam sets button's round rect corner, lParam is redraw flag
%RBM_SETBTNCOLOR   = %WM_USER + 2   'wParam sets button's face color, lParam is redraw flag
%RBM_SETBKMODE     = %WM_USER + 4   'wParam sets button's background mode, lParam is redraw flag
%RBM_SETBKCOLOR    = %WM_USER + 5   'wParam sets button's text background color, lParam is redraw flag
%RBM_SETTXTCOLOR   = %WM_USER + 6   'wParam sets button's text color, lParam is redraw flag
%RBM_SETTXTCOLORDN = %WM_USER + 7   'wParam sets button's text color in Down state, lParam is redraw flag
%RBM_SETTXTCOLORUP = %WM_USER + 8   'wParam sets button's text color in Up state, lParam is redraw flag
%RBM_SETTXT3D      = %WM_USER + 9   'wParam sets 3-D text on/off, lParam is redraw flag
%RBM_SETRAISED     = %WM_USER + 10  'wParam sets flat/raised appearence, lParam is redraw flag
%RBM_SETLABELMODE  = %WM_USER + 11  'wParam sets label mode, 0=none, 1=flat, 2=raised, 3=sunken, lParam is redraw flag
%RBM_SETCURSOR     = %WM_USER + 12  'with handle from resource, use lParam - else wParam can set system cursor (IDC_ ..)
%RBM_SETSHOWFOCUS  = %WM_USER + 13  'wParam sets focus rect on/off, lParam is redraw flag
%RBM_REFRESH       = %WM_USER + 14  'no params, simple InvalidateRect/UpdateWindow..
%RBM_SETTEXTPOS    = %WM_USER + 20  'wParam is DrawText position flags, lParam is redraw flag
'Note: see win32.hlp for more info in DrawText flags, for above message.
%RBM_SETTXTMARGINL = %WM_USER + 23  'wParam sets text margin Left,   lParam is redraw flag
%RBM_SETTXTMARGINT = %WM_USER + 24  'wParam sets text margin Top,    lParam is redraw flag
%RBM_SETTXTMARGINR = %WM_USER + 25  'wParam sets text margin Right,  lParam is redraw flag
%RBM_SETTXTMARGINB = %WM_USER + 26  'wParam sets text margin Bottom, lParam is redraw flag
%RBM_SETIMAGEPOS   = %WM_USER + 40  'if wParam < 0, table in WM_PAINT is used, else wParam/lParam sets x/y

' TABLE FOR wParam VALUES IN %RBM_SETIMAGEPOS:
'  -1 = LEFT, TOP
'  -2 = CENTERED, TOP
'  -3 = RIGHT, TOP
'  -4 = LEFT, CENTERED
'  -5 = CENTERED, CENTERED
'  -6 = RIGHT, CENTERED
'  -7 = LEFT, BOTTOM
'  -8 = CENTERED, BOTTOM
'  -9 = RIGHT, BOTTOM
'   0 or above = absolute X position in control, where lParam sets Y position

%RBM_SETIMAGESIZE  = %WM_USER + 41  'sets image size for DrawState (bitmap) and/or DrawIconEx
%RBM_SETIMGPOSLOCK = %WM_USER + 42  'wParam sets image position as locked on/off (no up/down drawn)
%RBM_SETIMGMARGINL = %WM_USER + 43  'wParam sets image margin Left,   lParam is redraw flag
%RBM_SETIMGMARGINT = %WM_USER + 44  'wParam sets image margin Top,    lParam is redraw flag
%RBM_SETIMGMARGINR = %WM_USER + 45  'wParam sets image margin Right,  lParam is redraw flag
%RBM_SETIMGMARGINB = %WM_USER + 46  'wParam sets image margin Bottom, lParam is redraw flag
%RBM_SETIMAGE      = %WM_USER + 47  'wParam is image type, bitmap or icon. lParam is resource or file name
%RBM_SETIMAGEUP    = %WM_USER + 48  'used to set separate image for button-up state. Else same as above
%RBM_SETIMAGEDN    = %WM_USER + 49  'used to set separate image for button-down state. Else same as above
%RBM_SETIMAGEDIS   = %WM_USER + 50  'used to set separate image for disabled state. Else same as above
%RBM_SETTOGGLE     = %WM_USER + 51  'wParam sets toggled value

%RBN_BTNUP         = %WM_USER + 230 'notification, triggered when button lifts
%RBN_BTNDN         = %WM_USER + 231 'notification, triggered when button is pressed down
%RBN_BTNFLAT       = %WM_USER + 232 'notification, triggered when button becomes flat again

%RBM_GETCORNER     = %WM_USER + 101 'returns button's round rect corner
%RBM_GETBTNCOLOR   = %WM_USER + 102 'returns button's face color
%RBM_GETBTNBRUSH   = %WM_USER + 103 'returns handle for button's color brush
%RBM_GETBKMODE     = %WM_USER + 104 'returns button's background mode
%RBM_GETBKCOLOR    = %WM_USER + 105 'returns button's text background color
%RBM_GETTXTCOLOR   = %WM_USER + 106 'returns button's text color
%RBM_GETTXTCOLORDN = %WM_USER + 107 'returns button's text color in Down state
%RBM_GETTXTCOLORUP = %WM_USER + 108 'returns button's text color in Up state
%RBM_GETTXT3D      = %WM_USER + 109 'returns 3-D text status
%RBM_GETRAISED     = %WM_USER + 110 'returns button's flat/raised flag
%RBM_GETLABELMODE  = %WM_USER + 111 'returns label mode, 0=none, 1=flat, 2=raised, 3=sunken
%RBM_GETSHOWFOCUS  = %WM_USER + 112 'returns button's focus rect on/off flag
%RBM_GETTEXTPOS    = %WM_USER + 120 'returns button's DrawText position flags
%RBM_GETTXTMARGINL = %WM_USER + 123 'returns button's text margin Left
%RBM_GETTXTMARGINT = %WM_USER + 124 'returns button's text margin Top
%RBM_GETTXTMARGINR = %WM_USER + 125 'returns button's text margin Right
%RBM_GETTXTMARGINB = %WM_USER + 126 'returns button's text margin Bottom
%RBM_GETIMAGEPOS   = %WM_USER + 140 'returns button's image position value
%RBM_GETIMGWIDTH   = %WM_USER + 141 'returns button's image width
%RBM_GETIMGHEIGHT  = %WM_USER + 142 'returns button's image height
%RBM_GETIMGMARGINL = %WM_USER + 143 'returns button's text margin Left
%RBM_GETIMGMARGINT = %WM_USER + 144 'returns button's text margin Top
%RBM_GETIMGMARGINR = %WM_USER + 145 'returns button's text margin Right
%RBM_GETIMGMARGINB = %WM_USER + 146 'returns button's text margin Bottom
%RBM_GETTOGGLED    = %WM_USER + 151 'returns toggled state

TYPE RRBUTTONDATA    'Following UDT can be freely customized to meet extended needs
  hWnd       AS DWORD 'button's handle
  id         AS LONG 'button's id
  hInst      AS DWORD 'button's instance handle
  hParent    AS DWORD 'button's parent's handle
  hBrush     AS DWORD 'button's face color brush
  hFont      AS DWORD 'button's font
  hRgn       AS DWORD 'button's round rect region, for redraw and clicks..
  hTimer     AS DWORD 'timer for tracking mouse vs. button states.
  corner     AS LONG 'button's round rect corner
  cursor     AS LONG 'default is IDC_ARROW
  btnColor   AS LONG 'button's face color
  bkMode     AS LONG 'button's text background mode, transparent or opaque
  bkColor    AS LONG 'button's text background color
  txtColor   AS LONG 'button's text color
  txtColorDn AS LONG 'button's text color in down state
  txtColorUp AS LONG 'button's text color in up-state
  txt3D      AS LONG '3-D text on/off
  raised     AS LONG 'button's flat/raised flag
  FocusRect  AS LONG 'button's focus rect on/off flag
  dn         AS LONG 'used internally, button's down state flag
  bClicked   AS LONG 'used internally, button's clicked flag
  toggled    AS LONG 'stores toggled value
  labelMode  AS LONG 'stores label mode, 0=none (button), 1=flat, 2=raised, 3=sunken
  hImg       AS DWORD 'image handle for button flat state
  hImgUp     AS DWORD 'image handle for button up state
  hImgDn     AS DWORD 'image handle for button down state
  hImgDis    AS DWORD 'image handle for disabled button's image
  imgHasPnt  AS LONG 'tells if image is handled by parent (via BM_SETIMAGE)
  imgType    AS LONG 'bmp or icon?
  imgX       AS LONG 'button's image Left position
  imgY       AS LONG 'button's image Top position
  imgW       AS LONG 'button's image Width
  imgH       AS LONG 'button's image Height
  imgPos     AS LONG 'button's image position flag, for table in WM_PAINT
  imgPosLock AS LONG 'button's image position locked flag
  imgMarginL AS LONG 'button's image margin Left
  imgMarginT AS LONG 'button's image margin Top
  imgMarginR AS LONG 'button's image margin Right
  imgMarginB AS LONG 'button's image margin Bottom
  txtPos     AS LONG 'button's text position flag, for DrawText
  txtMarginL AS LONG 'button's text margin Left
  txtMarginT AS LONG 'button's text margin Top
  txtMarginR AS LONG 'button's text margin Right
  txtMarginB AS LONG 'button's text margin Bottom
  PalClr(5)  AS LONG 'array for some colors used by the control
END TYPE

DECLARE SUB CreateRRButtonPalette(BYVAL hWnd AS DWORD)
DECLARE SUB DrawRRButton(BYVAL hWnd AS DWORD, BYVAL memDC AS DWORD, rc AS RECT)
DECLARE FUNCTION CursorPtInRegion(BYVAL hWnd AS DWORD) AS LONG
DECLARE FUNCTION CreateRRButton(BYVAL hParent AS DWORD, BYVAL id AS LONG, BYVAL txt AS STRING, _
                                BYVAL vLeft AS LONG, BYVAL vTop AS LONG, _
                                BYVAL vWidth AS LONG, BYVAL vHeight AS LONG, _
                                BYVAL wStyle AS LONG, BYVAL wStyleEx AS LONG, _
                                BYVAL DDTstyle AS LONG) AS DWORD
DECLARE FUNCTION RRButtonProc (BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                               BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG

'
' Create RRBUTTON control
'
FUNCTION CreateRRButton(BYVAL hParent AS DWORD, BYVAL id AS LONG, BYVAL txt AS STRING, _
                        BYVAL vLeft AS LONG, BYVAL vTop AS LONG, _
                        BYVAL vWidth AS LONG, BYVAL vHeight AS LONG, _
                        BYVAL wStyle AS LONG, BYVAL wStyleEx AS LONG, _
                        BYVAL DDTstyle AS LONG) AS DWORD

  LOCAL hBtn        AS DWORD
  LOCAL wc          AS WNDCLASSEX
  LOCAL szClassName AS ASCIIZ * 10

  szClassName      = "RRBUTTON"
  IF GetClassInfoEx(GetModuleHandle(BYVAL %NULL), szClassName, wc) = 0 THEN
     wc.cbSize        = SIZEOF(wc)
     wc.style         = %CS_HREDRAW OR %CS_VREDRAW OR %CS_DBLCLKS
     wc.lpfnWndProc   = CODEPTR(RRButtonProc)
     wc.cbWndExtra    = 4 'for pointer to RRBUTTONDATA structure
     wc.hInstance     = GetWindowLong(hParent, %GWL_HINSTANCE)
     wc.lpszClassName = VARPTR(szClassName)

     CALL RegisterClassEx(wc)
     IF ERR THEN EXIT FUNCTION         'something went wrong, so bail out
  END IF

  IF DDTstyle THEN 'create control using dialog units measurements (DDT)
     CONTROL ADD "RRBUTTON", hParent, id, txt, vLeft, vTop, vWidth, vHeight, wStyle, wStyleEx
     CONTROL HANDLE hParent, id TO hBtn

  ELSE             'create control using pixels measurements (SDK/API..)
     hBtn = CreateWindowEx(wStyleEx, "RRBUTTON", BYVAL STRPTR(txt), wStyle, _
                             vLeft, vTop, vWidth, vHeight, _
                             hParent, id, GetWindowLong(hParent, %GWL_HINSTANCE), BYVAL 0)
  END IF

  FUNCTION = hBtn 'return handle
END FUNCTION

'
' RRBUTTON Control message handling procedure
'------------------------------------------------------------------------------
FUNCTION RRButtonProc (BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                       BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG

  LOCAL rr AS RRBUTTONDATA PTR 'get control specific data, if possible
  IF hWnd AND wMsg <> %WM_CREATE THEN rr = GetWindowLong(hWnd, 0)

  SELECT CASE wMsg
     CASE %WM_CREATE
        rr = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, SIZEOF(@rr))
        IF rr THEN
           SetWindowLong hWnd, 0, rr  'Store the pointer for later use
        ELSE
           FUNCTION = -1 : EXIT FUNCTION  'Return -1 to break the action
        END IF

        LOCAL hDC AS DWORD, memDC AS DWORD, hbit AS DWORD, hImg AS DWORD, hBrush AS DWORD
        LOCAL ix AS LONG, iy AS LONG, imgLoad AS LONG, tmpDn AS LONG
        LOCAL zPtr AS ASCIIZ PTR, zCap AS ASCIIZ * 1024
        LOCAL ps   AS PaintStruct
        LOCAL rc   AS RECT, rcTx AS RECT, rcUp AS RECT, rcDn AS RECT, rcDis AS RECT

        @rr.hParent     = GetWindowLong(hWnd, %GWL_HWNDPARENT)
        @rr.hWnd        = hWnd
        @rr.id          = GetWindowLong(hWnd, %GWL_ID)
        @rr.hInst       = GetWindowLong(hWnd, %GWL_HINSTANCE)
        @rr.corner      = 20
        @rr.cursor      = LoadCursor(%NULL, BYVAL %IDC_ARROW)
        @rr.bkMode      = %TRANSPARENT
        @rr.PalClr(5)   = -1
        IF GetSysColor(%COLOR_BTNFACE) = &H0& OR _
           GetSysColor(%COLOR_BTNFACE) = &HFFFFFF& THEN 'if black or white system button - high contrast?
           @rr.btnColor = GetSysColor(%COLOR_BTNFACE)
        ELSE
           @rr.btnColor    = GetSysColor(%COLOR_BTNFACE) + &H080808
        END IF
        @rr.bkColor     = @rr.btnColor
        @rr.txtColor    = GetSysColor(%COLOR_BTNTEXT)
        @rr.txtColorUp  = @rr.txtColor
        @rr.txtColorDn  = @rr.txtColor
        @rr.raised      = 0
        @rr.labelMode   = 0  '0=none (button), 1=flat label, 2=raised, 3=sunken
        @rr.FocusRect   = 1
        @rr.imgPos      = -4                        'left centered
        @rr.txtPos      = %DT_SINGLELINE OR %DT_VCENTER OR %DT_CENTER 'centered centered
        @rr.imgW        = 16                        'assign 16x16 image size at start
        @rr.imgH        = 16

        SendMessage hWnd, %RBM_SETCORNER, @rr.corner, 0   'initiate region
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_COMMAND, %WM_NOTIFY 'in case button is used as container for other controls..
         SendMessage @rr.hParent, wMsg, wParam, lParam

     CASE %WM_DESTROY
        IF @rr.hTimer    THEN KillTimer hWnd, @rr.id
        IF @rr.PalClr(0) THEN DeleteObject @rr.PalClr(0)
        IF @rr.hBrush    THEN DeleteObject @rr.hBrush
        IF @rr.hRgn      THEN DeleteObject @rr.hRgn
        IF @rr.hImg AND @rr.imgHasPnt = 0 THEN DeleteObject @rr.hImg
        IF @rr.hImgUp    THEN DeleteObject @rr.hImgUp
        IF @rr.hImgDn    THEN DeleteObject @rr.hImgDn
        IF @rr.hImgDis   THEN DeleteObject @rr.hImgDis
        IF rr THEN CALL HeapFree(GetProcessHeap(), 0, BYVAL rr)
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_ENABLE
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_KILLFOCUS                    'Reset focus rect on button through WM_PAINT
        @rr.dn = 0 : @rr.bClicked = 0
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_SETFOCUS                        'Set focus rect on button through WM_PAINT
        @rr.dn = 0
        IF @rr.labelMode > 0 AND @rr.FocusRect = 0 THEN  'jump past labels, if button been
           IF (GetKeyState(%VK_SHIFT) AND &H8000) THEN   'changed to label and has WM_TABSTOP style
              SetFocus GetNextDlgTabItem(@rr.hParent, hWnd, %TRUE)
           ELSE
              SetFocus GetNextDlgTabItem(@rr.hParent, hWnd, %FALSE)
           END IF
        END IF
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_GETDLGCODE 'MS system responds irregular to Enter key - better handle it ourselves
        IF @rr.labelMode = 0 OR @rr.FocusRect <> 0 THEN
           LOCAL pMsg AS TAGMSG PTR, iMsg AS LONG

           pMsg = lParam
           IF pMsg > 0 THEN
              iMsg = @pMsg.Message
              IF iMsg = %WM_KEYDOWN OR iMsg = %WM_CHAR THEN 'all we want to handle ourselves is Enter key, so..
                 IF @pMsg.wParam = %VK_RETURN THEN
                    FUNCTION = %DLGC_WANTALLKEYS
                    EXIT FUNCTION
                 END IF
              END IF
           END IF

           ' following enables mnemorics, i.e, button is triggered
           ' when the character that follows ampersand (&) is pressed
           IF lParam = 0 AND @rr.labelMode = 0 THEN
              FUNCTION = %DLGC_BUTTON OR %DLGC_UNDEFPUSHBUTTON
              EXIT FUNCTION
           END IF
        END IF

     CASE %WM_SYSCOLORCHANGE
        IF GetSysColor(%COLOR_BTNFACE) = &H0& OR _
           GetSysColor(%COLOR_BTNFACE) = &HFFFFFF& THEN 'if black or white system button - high contrast?
           SendMessage hWnd, %RBM_SETBTNCOLOR, GetSysColor(%COLOR_BTNFACE), 0
           SendMessage hWnd, %RBM_SETTXTCOLOR, GetSysColor(%COLOR_BTNTEXT), 0
           SendMessage hWnd, %RBM_SETTXTCOLORUP, GetSysColor(%COLOR_BTNTEXT), 0
           SendMessage hWnd, %RBM_SETTXTCOLORDN, GetSysColor(%COLOR_BTNTEXT), 1
        END IF

     CASE %WM_KEYDOWN                      'we handle enter and space keys ourself
        IF @rr.bClicked THEN EXIT SELECT   'avoid action while mouse button down
        SELECT CASE wParam
           CASE %VK_RETURN
              @rr.toggled = NOT @rr.toggled
              CALL SendMessage(@rr.hParent, %WM_COMMAND, MAKLNG(@rr.id, %BN_CLICKED), hWnd)

           CASE %VK_SPACE
              @rr.dn = 2 : @rr.bClicked = 2
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
        END SELECT

     CASE %WM_KEYUP
        IF @rr.bClicked = 2 THEN
           @rr.bClicked = 0 ' <- flag for %WM_MOUSEMOVE
           IF @rr.dn THEN
              @rr.dn = 0
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
              IF IsWindowEnabled(hWnd) THEN
                 @rr.toggled = NOT @rr.toggled
                 SendMessage @rr.hParent, %WM_COMMAND, _
                             MAKLNG(@rr.id, %BN_CLICKED), hWnd
              END IF
           END IF
           @rr.dn = 0
           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
        END IF

     CASE %WM_LBUTTONDBLCLK 'forward this for rapid button clicking...
        CALL SendMessage(hWnd,%WM_LBUTTONDOWN,wParam,lParam)
        FUNCTION = 0 :EXIT FUNCTION

     CASE %WM_LBUTTONDOWN
        IF CursorPtInRegion(hWnd) THEN
           SetCapture hWnd
           IF @rr.labelMode > 0 AND @rr.FocusRect = 0 THEN EXIT FUNCTION
           IF @rr.bClicked = 2 THEN EXIT SELECT
           @rr.bClicked = 1                        '<- flag for %WM_MOUSEMOVE
           IF @rr.labelMode = 0 OR (@rr.labelMode > 0 AND @rr.FocusRect = 1) THEN 'if button, or label set to show focusrect
              IF GetFocus <> hWnd THEN SetFocus hWnd     'set focus to clicked control
           END IF
           @rr.dn = 2
           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
           FUNCTION = 0 : EXIT FUNCTION
        END IF

     CASE %WM_LBUTTONUP
        ReleaseCapture
        IF @rr.labelMode > 0 AND @rr.FocusRect = 0 THEN EXIT FUNCTION
        IF @rr.bClicked = 0 THEN EXIT FUNCTION
        tmpDn = @rr.dn

        IF CursorPtInRegion(hWnd) THEN
           @rr.dn = 1 : @rr.bClicked = 0
           IF @rr.labelMode THEN @rr.dn = 0
        ELSE
           @rr.dn = 0 : @rr.bClicked = 0
        END IF
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd

        IF tmpDn AND IsWindowEnabled(hWnd) THEN
           @rr.toggled = NOT @rr.toggled
           SendMessage @rr.hParent, %WM_COMMAND, MAKLNG(@rr.id, %BN_CLICKED), hWnd
        END IF
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_MOUSEMOVE
        IF @rr.labelMode > 0 AND @rr.FocusRect = 0 THEN
           IF CursorPtInRegion(hWnd) THEN
              IF GetCursor() <> @rr.cursor THEN SetCursor @rr.cursor
           END IF
           EXIT FUNCTION
        END IF

        IF @rr.bClicked = 2 THEN EXIT SELECT

        IF CursorPtInRegion(hWnd) THEN
           IF GetCursor() <> @rr.cursor THEN SetCursor @rr.cursor
           IF @rr.hTimer = 0 AND @rr.labelMode = 0 THEN
              @rr.hTimer = SetTimer(hWnd, @rr.id, 10, 0)
           END IF
           IF @rr.bClicked THEN
              IF @rr.dn = 2 THEN EXIT SELECT
              @rr.dn = 2
           ELSE
              IF @rr.dn = 1 THEN EXIT SELECT
              @rr.dn = 1
              IF @rr.labelMode THEN @rr.dn = 0
           END IF
        ELSE
           IF GetCursor() = @rr.cursor THEN
              SetCursor LoadCursor(%NULL, BYVAL %IDC_ARROW)
           END IF
           IF @rr.dn = 0 THEN EXIT SELECT
           @rr.dn = 0
        END IF
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd

     CASE %WM_TIMER
        IF CursorPtInRegion(hWnd) = 0 THEN
           KillTimer hWnd, @rr.id
           @rr.hTimer = 0
           @rr.dn = 0
           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hWnd
        END IF

     CASE %WM_ERASEBKGND
        FUNCTION = 1 : EXIT FUNCTION

     CASE %WM_PAINT
        'Individual controls do not receive WM_SYSCOLORCHANGE message
        'from stupid Windows when/if a user for example changes the
        'color scheme. By using a reference color, following ensures
        'the control updates brush + palette to correct colors anyway.
        IF GetSysColor(%COLOR_BTNFACE) <> @rr.PalClr(5) THEN
           CreateRRButtonPalette hWnd
        END IF

        '=== Get Rect and Caption ==================================
        CALL GetClientRect(hWnd, rc)
        CALL GetWindowText(hWnd, zCap, 1024)

        '=== Initialize ============================================
        BeginPaint hWnd, ps
        IF @rr.hRgn THEN SelectObject ps.hDC, @rr.hRgn

        IF @rr.PalClr(0) THEN 'if we have palette (256 color mode), then use it DC..
           SelectPalette ps.hDC, @rr.PalClr(0), 1
           RealizePalette ps.hDC
        END IF

        memDC = CreateCompatibleDC(ps.hDC)

        IF @rr.PalClr(0) THEN 'if we have palette (256 color mode), then use it memDC..SelectClipRgn
           SelectPalette memDC, @rr.PalClr(0), 1
           RealizePalette memDC
        END IF

        hbit  = CreateCompatibleBitmap(ps.hDC, rc.nRight, rc.nBottom)
        IF hbit      THEN hbit = SelectObject(memDC, hbit)
        IF @rr.hRgn  THEN SelectObject memDC, @rr.hRgn
        IF @rr.hFont THEN @rr.hFont = SelectObject(memDC, @rr.hFont)

        CALL SetBkMode(memDC, @rr.bkMode)
        CALL SetBkColor(memDC, @rr.bkColor)
        FillRect memDC, rc, GetSysColorBrush(%COLOR_BTNFACE) 'cls entire memDC

        'No %WM_CTLCOLORBTN - we use custom messages to change color/text color instead.
        hBrush = SelectObject(memDC, @rr.hBrush)

        '=== DETERMINE IMAGE POS ========
        IF @rr.hImg THEN
           SELECT CASE @rr.imgPos
           CASE -1                                              'LEFT, TOP
              ix = @rr.imgMarginL                                 'left align
              iy = @rr.imgMarginT                                 'top align
           CASE -2                                              'CENTERED, TOP
              ix = @rr.imgMarginL + (rc.nRight - @rr.imgW) / 2    'center x
              iy = @rr.imgMarginT                                 'top align
           CASE -3                                              'RIGHT, TOP
              ix = rc.nRight - @rr.imgW - @rr.imgMarginR          'right align
              iy = @rr.imgMarginT                                 'top align
           CASE -4                                              'LEFT, CENTERED
              ix = @rr.imgMarginL                                 'left align
              iy = @rr.imgMarginT + (rc.nBottom - @rr.imgH) / 2   'center y
           CASE -5                                              'CENTERED, CENTERED
              ix = @rr.imgMarginL + (rc.nRight - @rr.imgW) / 2    'center x
              iy = @rr.imgMarginT + (rc.nBottom - @rr.imgH) / 2   'center y
           CASE -6                                              'RIGHT, CENTERED
              ix = rc.nRight - @rr.imgW - @rr.imgMarginR          'right align
              iy = @rr.imgMarginT + (rc.nBottom - @rr.imgH) / 2   'center y
           CASE -7                                              'LEFT, BOTTOM
              ix = @rr.imgMarginL                                 'left align
              iy = rc.nBottom - @rr.imgH - @rr.imgMarginB         'bottom align
           CASE -8                                              'CENTERED, BOTTOM
              ix = @rr.imgMarginL + (rc.nRight - @rr.imgW) / 2    'center x
              iy = rc.nBottom - @rr.imgH - @rr.imgMarginB         'bottom align
           CASE -9                                              'RIGHT, BOTTOM
              ix = rc.nRight - @rr.imgW - @rr.imgMarginR          'right align
              iy = rc.nBottom - @rr.imgH - @rr.imgMarginB         'bottom align
           CASE ELSE
              ix = @rr.imgX                                     'ELSE ABSOLUTE X
              iy = @rr.imgY                                     '     ABSOLUTE Y
           END SELECT
        END IF
        rcTx.nLeft =  @rr.txtMarginL : rcTx.nRight  = rc.nRight  - @rr.txtMarginR
        rcTx.nTop  =  @rr.txtMarginT : rcTx.nBottom = rc.nBottom - @rr.txtMarginB
        rcUp.nLeft = rcTx.nLeft     : rcUp.nRight  = rcTx.nRight
        rcUp.nTop  = rcTx.nTop      : rcUp.nBottom = rcTx.nBottom
        rcDn.nLeft = rcTx.nLeft + 1 : rcDn.nRight  = rcTx.nRight  + 1
        rcDn.nTop  = rcTx.nTop  + 1 : rcDn.nBottom = rcTx.nBottom + 1
        IF @rr.labelMode THEN
           DECR rcDn.nLeft : DECR rcDn.nRight
           DECR rcDn.nTop  : DECR rcDn.nBottom
        END IF

        '=== Draw Button =====================================
        DrawRRButton hWnd, memDC, rc

        IF @rr.dn = 2 THEN     'BUTTON DOWN
           IF @rr.txt3D THEN
              rcDis.nLeft = rcDn.nLeft + 1 : rcDis.nRight  = rcDn.nRight  + 1
              rcDis.nTop  = rcDn.nTop + 1  : rcDis.nBottom = rcDn.nBottom + 1
              CALL SetTextColor(memDC, GetSysColor(%COLOR_3DHILIGHT))
              DrawText memDC, zCap, -1, rcDis, @rr.txtPos
           END IF

           CALL SetTextColor(memDC, @rr.txtColorDn)
           DrawText memDC, zCap, -1, rcDn, @rr.txtPos
           IF @rr.hImgDn THEN
              hImg = @rr.hImgDn
           ELSEIF @rr.hImgUp THEN
              hImg = @rr.hImgUp
           ELSE
              hImg = @rr.hImg
           END IF
           IF hImg THEN
              IF @rr.imgType = %IMAGE_BITMAP THEN
                 IF @rr.labelMode = 0 AND @rr.imgPosLock = 0 THEN
                    CALL DrawState(memDC, 0&, 0&, hImg, 0&, ix + 1, iy + 1, 0, 0, &H4)
                 ELSE
                    CALL DrawState(memDC, 0&, 0&, hImg, 0&, ix, iy, 0, 0, &H4)
                 END IF
              ELSE
                 IF @rr.labelMode = 0 AND @rr.imgPosLock = 0 THEN
                    DrawIconEx memDC, ix + 1, iy + 1, hImg, @rr.imgW, @rr.imgH, 0, BYVAL 0, %DI_NORMAL
                 ELSE
                    DrawIconEx memDC, ix, iy, hImg, @rr.imgW, @rr.imgH, 0, BYVAL 0, %DI_NORMAL
                 END IF
              END IF
           END IF

        ELSEIF @rr.dn = 1 THEN 'BUTTON UP
           'DECR rcUp.nTop : DECR rcUp.nBottom : DECR rcUp.nLeft : DECR rcUp.nRight
           IF @rr.txt3D THEN
              rcDis.nLeft = rcUp.nLeft + 1 : rcDis.nRight  = rcUp.nRight  + 1
              rcDis.nTop  = rcUp.nTop + 1  : rcDis.nBottom = rcUp.nBottom + 1
              CALL SetTextColor(memDC, GetSysColor(%COLOR_3DHILIGHT))
              DrawText memDC, zCap, -1, rcDis, @rr.txtPos
           END IF

           CALL SetTextColor(memDC, @rr.txtColorUp)
           DrawText memDC, zCap, -1, rcUp, @rr.txtPos
           IF @rr.hImgUp THEN
              hImg = @rr.hImgUp
           ELSEIF @rr.hImgDn THEN
              hImg = @rr.hImgDn
           ELSE
              hImg = @rr.hImg
           END IF
           IF hImg THEN
              IF @rr.imgType = %IMAGE_BITMAP THEN
                 CALL DrawState(memDC, 0&, 0&, hImg, 0&, ix, iy, 0, 0, &H4)
              ELSE
                 DrawIconEx memDC, ix, iy, hImg, @rr.imgW, @rr.imgH, 0, BYVAL 0, %DI_NORMAL
              END IF
           END IF

        ELSE                   'FLAT MODE
           IF IsWindowEnabled(hWnd) = 0 THEN 'if Button is disabled
              rcDis.nLeft = rcTx.nLeft + 1 : rcDis.nRight  = rcTx.nRight  + 1
              rcDis.nTop  = rcTx.nTop + 1  : rcDis.nBottom = rcTx.nBottom + 1
              CALL SetTextColor(memDC, GetSysColor(%COLOR_3DHILIGHT))
              DrawText memDC, zCap, -1, rcDis, @rr.txtPos
              CALL SetTextColor(memDC, GetSysColor(%COLOR_3DSHADOW))

           ELSE
              IF @rr.txt3D THEN
                 rcDis.nLeft = rcTx.nLeft + 1 : rcDis.nRight  = rcTx.nRight  + 1
                 rcDis.nTop  = rcTx.nTop + 1  : rcDis.nBottom = rcTx.nBottom + 1
                 CALL SetTextColor(memDC, GetSysColor(%COLOR_3DHILIGHT))
                 DrawText memDC, zCap, -1, rcDis, @rr.txtPos
              END IF

              CALL SetTextColor(memDC, @rr.txtColor)
           END IF
           DrawText memDC, zCap, -1, rcTx, @rr.txtPos

           IF @rr.hImg THEN
              IF IsWindowEnabled(hWnd) THEN
                 IF @rr.imgType = %IMAGE_BITMAP THEN
                    CALL DrawState(memDC, 0, 0, @rr.hImg, 0, ix, iy, 0, 0, &H4)
                 ELSE
                    DrawIconEx memDC, ix, iy, @rr.hImg, @rr.imgW, @rr.imgH, 0, BYVAL 0, %DI_NORMAL
                 END IF

              ELSE
                 IF @rr.hImgDis THEN
                    hImg = @rr.hImgDis
                 ELSE
                    hImg = @rr.hImg
                 END IF

                 IF @rr.imgType = %IMAGE_BITMAP THEN
                    CALL DrawState(memDC, 0, 0, hImg, 0, ix, iy, 0, 0, &H4)
                 ELSE
                    DrawIconEx memDC, ix, iy, hImg, @rr.imgW, @rr.imgH, 0, BYVAL 0, %DI_NORMAL
                 END IF
              END IF
           END IF

        END IF

        BitBlt ps.hDC, 0, 0, rc.nright, rc.nbottom, memDC, 0, 0, %SRCCOPY 'paint it all

        '=== End paint procedure ====================================
        IF @rr.hFont THEN @rr.hFont = SelectObject(memDC, @rr.hFont)
        IF hBrush    THEN SelectObject memDC, hBrush
        IF hbit  THEN DeleteObject SelectObject(memDC, hbit)
        IF memDC THEN DeleteDC memDC
        EndPaint hWnd, ps
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_SETFONT
        IF wParam THEN                             'if we get something
           @rr.hFont = wParam                      'store identifier of font
           IF lParam  AND IsWindowVisible(hWnd) THEN 'if redraw flag is set and visible
              InvalidateRect hWnd, BYVAL %NULL, 0  'redraw button
              UpdateWindow hWnd
           END IF
        END IF

     CASE %WM_SIZE
        CALL GetClientRect(hWnd, rc)               'get button size and create new region
        hDC = GetDc(hWnd)
        FillRect hDC, rc, GetSysColorBrush(%COLOR_BTNFACE)
        ReleaseDc hWnd, hDC

        IF @rr.hRgn THEN DeleteObject @rr.hRgn     'destroy old region
        @rr.hRgn = CreateRoundRectRgn(0, 0, rc.nRight + 1, rc.nBottom + 1, @rr.corner, @rr.corner)
        FUNCTION = 0 : EXIT FUNCTION

'------------------------------------------------------------------------------
' Control specific messages
'------------------------------------------------------------------------------
     CASE %RBM_REFRESH
        InvalidateRect hWnd, BYVAL %NULL, 0        'redraw button
        UpdateWindow hWnd
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETBTNCOLOR
        @rr.btnColor = wParam                     'wParam sets new color
        CreateRRButtonPalette hWnd
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0

     CASE %RBM_SETCORNER
        CALL GetClientRect(hWnd, rc)                          'get button's size
        hDC = GetDc(hWnd)
        FillRect hDC, rc, GetSysColorBrush(%COLOR_BTNFACE)    'wipe entire button area
        ReleaseDc hWnd, hDC

        @rr.corner = MAX&(0, wParam)                          'store new value (never negative)
        IF @rr.hRgn THEN DeleteObject @rr.hRgn                'destroy old region and create new
        @rr.hRgn = CreateRoundRectRgn(0, 0, rc.nRight + 1, rc.nBottom + 1, @rr.corner, @rr.corner)
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0   'lParam decides redraw
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETRAISED
        IF wParam <> 0 THEN wParam = %TRUE  'correct eventual mistakes
        @rr.raised = wParam                 'store setting
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETLABELMODE  'wParam sets label mode, '0=none (button), 1=flat label, 2=raised, 3=sunken
        @rr.labelMode = wParam
        IF wParam THEN @rr.FocusRect = 0 'so labels don't show focus rect as default
        IF @rr.labelMode THEN @rr.dn = @rr.labelMode - 1 ELSE @rr.dn = 0
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETCURSOR
        IF lParam THEN
           @rr.cursor = lParam
        ELSE
           @rr.cursor = LoadCursor(%NULL, BYVAL wParam)
        END IF

     CASE %RBM_SETSHOWFOCUS
        @rr.FocusRect = wParam                'store setting
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_GETCORNER    : FUNCTION = @rr.corner    : EXIT FUNCTION
     CASE %RBM_GETRAISED    : FUNCTION = @rr.raised    : EXIT FUNCTION
     CASE %RBM_GETLABELMODE : FUNCTION = @rr.labelMode : EXIT FUNCTION
     CASE %RBM_GETSHOWFOCUS : FUNCTION = @rr.FocusRect : EXIT FUNCTION

'TEXT MESSAGES
     CASE %RBM_SETTEXTPOS
        @rr.txtPos = wParam
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETBKMODE
        @rr.bkMode = wParam
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETBKCOLOR
        @rr.bkColor = wParam
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETTXTCOLOR
        @rr.txtColor = wParam
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETTXTCOLORDN
        @rr.txtColorDn = wParam
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETTXTCOLORUP
        @rr.txtColorUp = wParam
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETTXT3D
        @rr.txt3D = wParam
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_SETTXTMARGINL TO %RBM_SETTXTMARGINB            'set text margins
        SELECT CASE wMsg
           CASE %RBM_SETTXTMARGINL : @rr.txtMarginL = wParam 'left
           CASE %RBM_SETTXTMARGINR : @rr.txtMarginR = wParam 'right
           CASE %RBM_SETTXTMARGINT : @rr.txtMarginT = wParam 'top
           CASE %RBM_SETTXTMARGINB : @rr.txtMarginB = wParam 'bottom
        END SELECT
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0
        FUNCTION = 0 : EXIT FUNCTION

     CASE %RBM_GETTEXTPOS    : FUNCTION = @rr.txtPos     : EXIT FUNCTION
     CASE %RBM_GETBKMODE     : FUNCTION = @rr.bkMode     : EXIT FUNCTION
     CASE %RBM_GETBKCOLOR    : FUNCTION = @rr.bkColor    : EXIT FUNCTION
     CASE %RBM_GETTXTCOLOR   : FUNCTION = @rr.txtColor   : EXIT FUNCTION
     CASE %RBM_GETTXTCOLORDN : FUNCTION = @rr.txtColorDn : EXIT FUNCTION
     CASE %RBM_GETTXTCOLORUP : FUNCTION = @rr.txtColorUp : EXIT FUNCTION
     CASE %RBM_GETTXT3D      : FUNCTION = @rr.txt3D      : EXIT FUNCTION

     CASE %RBM_GETTXTMARGINL : FUNCTION = @rr.txtMarginL : EXIT FUNCTION
     CASE %RBM_GETTXTMARGINT : FUNCTION = @rr.txtMarginT : EXIT FUNCTION
     CASE %RBM_GETTXTMARGINR : FUNCTION = @rr.txtMarginR : EXIT FUNCTION
     CASE %RBM_GETTXTMARGINB : FUNCTION = @rr.txtMarginB : EXIT FUNCTION

'IMAGE MESSAGES
     CASE %BM_SETIMAGE              'use BM_SETIMAGE if parent maintains handle of image object
        @rr.imgType = wParam        'image type flag, %IMAGE_BITMAP or %IMAGE_ICON
        IF @rr.hImg THEN            'if we already have an image handle
           FUNCTION = @rr.hImg      'return the old handle
        ELSE
           FUNCTION = %NULL         'otherwise return null
        END IF
        @rr.hImg      = lParam      'image handle
        @rr.imgHasPnt = 1           'image is handled by parent

        IF @rr.hImg THEN
           @rr.imgW = 16            'assign 16x16 as default
           @rr.imgH = 16
        ELSE
           @rr.imgW = 0              'else no image
           @rr.imgH = 0
        END IF
        EXIT FUNCTION

     CASE %RBM_SETIMAGE              'use RBM_SETIMAGE if control is to maintain image object
        IF lParam THEN
           @rr.imgType = wParam                             'image type flag, %IMAGE_BITMAP or %IMAGE_ICON
           zPtr        = lParam                             'lParam is pointer to file- or resource name
           imgLoad     = %LR_LOADMAP3DCOLORS                'initial LoadImage flag
           IF INSTR(@zPtr, ".") THEN imgLoad = imgLoad OR %LR_LOADFROMFILE 'is file name, not resource (?)

           IF @rr.hImg THEN DeleteObject @rr.hImg
           @rr.hImg = LoadImage(@rr.hInst, @zPtr, wParam, 0, 0, imgLoad)

           IF @rr.hImg THEN
              @rr.imgW = 16          'set 16x16 as default
              @rr.imgH = 16
           ELSE
              @rr.imgW = 0           'else no image
              @rr.imgH = 0
           END IF
        ELSE
           IF @rr.hImg THEN DeleteObject @rr.hImg
           @rr.hImg = 0
        END IF

     'possible to set separate images for button-up, down and disabled state
     CASE %RBM_SETIMAGEUP, %RBM_SETIMAGEDN, %RBM_SETIMAGEDIS
        IF lParam THEN
           @rr.imgType = wParam     'image type flag, %IMAGE_BITMAP or %IMAGE_ICON
           zPtr        = lParam     'pointer to image resource/file name
           imgLoad     = %LR_LOADMAP3DCOLORS
           IF INSTR(@zPtr, ".") THEN imgLoad = imgLoad OR %LR_LOADFROMFILE 'is file name, not resource (?)

           IF wMsg = %RBM_SETIMAGEUP THEN
              IF @rr.hImgUp THEN DeleteObject @rr.hImgUp
              @rr.hImgUp = LoadImage(@rr.hInst, @zPtr, wParam, 0, 0, imgLoad)
           ELSEIF wMsg = %RBM_SETIMAGEDN THEN
              IF @rr.hImgDn THEN DeleteObject @rr.hImgDn
              @rr.hImgDn = LoadImage(@rr.hInst, @zPtr, wParam, 0, 0, imgLoad)
           ELSEIF wMsg = %RBM_SETIMAGEDIS THEN
              IF @rr.hImgDis THEN DeleteObject @rr.hImgDis
              @rr.hImgDis = LoadImage(@rr.hInst, @zPtr, wParam, 0, 0, imgLoad)
           END IF
        ELSE
           IF wMsg = %RBM_SETIMAGEUP THEN
              IF @rr.hImgUp THEN DeleteObject @rr.hImgUp
              @rr.hImgUp = 0
           ELSEIF wMsg = %RBM_SETIMAGEDN THEN
              IF @rr.hImgDn THEN DeleteObject @rr.hImgDn
              @rr.hImgDn = 0
           ELSEIF wMsg = %RBM_SETIMAGEDIS THEN
              IF @rr.hImgDis THEN DeleteObject @rr.hImgDis
              @rr.hImgDis = 0
           END IF

        END IF

     CASE %RBM_SETIMAGEPOS
        IF wParam < 0 THEN          'if negative
           @rr.imgPos = wParam      'table in WM_PAINT decides
        ELSE                        'else, if zero or positive
           @rr.imgPos = 0           'tell table in WM_PAINT we use absolute position
           @rr.imgX   = wParam      'absolute X pos
           @rr.imgY   = lParam      'absolute y pos
        END IF

     CASE %RBM_SETIMAGESIZE
        @rr.imgW = wParam            'set new image size, WM_PAINT will stretch accordingly
        @rr.imgH = lParam

     CASE %RBM_SETIMGPOSLOCK
        @rr.imgPosLock = wParam      'set image poition as locked - no up/down

     CASE %RBM_SETIMGMARGINL TO %RBM_SETIMGMARGINB            'set image margins
        SELECT CASE wMsg
           CASE %RBM_SETIMGMARGINL : @rr.imgMarginL = wParam 'left
           CASE %RBM_SETIMGMARGINR : @rr.imgMarginR = wParam 'right
           CASE %RBM_SETIMGMARGINT : @rr.imgMarginT = wParam 'top
           CASE %RBM_SETIMGMARGINB : @rr.imgMarginB = wParam 'bottom
        END SELECT
        IF lParam THEN SendMessage hWnd, %RBM_REFRESH, 0, 0

     CASE %RBM_GETBTNBRUSH   : FUNCTION = @rr.hBrush     : EXIT FUNCTION
     CASE %RBM_GETBTNCOLOR   : FUNCTION = @rr.btnColor   : EXIT FUNCTION
     CASE %RBM_GETIMAGEPOS   : FUNCTION = @rr.imgPos     : EXIT FUNCTION
     CASE %RBM_GETIMGWIDTH   : FUNCTION = @rr.imgW       : EXIT FUNCTION
     CASE %RBM_GETIMGHEIGHT  : FUNCTION = @rr.imgH       : EXIT FUNCTION
     CASE %RBM_GETIMGMARGINL : FUNCTION = @rr.imgMarginL : EXIT FUNCTION
     CASE %RBM_GETIMGMARGINT : FUNCTION = @rr.imgMarginT : EXIT FUNCTION
     CASE %RBM_GETIMGMARGINR : FUNCTION = @rr.imgMarginR : EXIT FUNCTION
     CASE %RBM_GETIMGMARGINB : FUNCTION = @rr.imgMarginB : EXIT FUNCTION

  END SELECT

  '* Default processing for other messages.
  FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)

END FUNCTION

'
' create palette for button, when  needed
'------------------------------------------------------------------------------
SUB CreateRRButtonPalette(BYVAL hWnd AS DWORD)
  LOCAL ii AS LONG, hDC AS DWORD, rr AS RRBUTTONDATA PTR
  rr = GetWindowLong(hWnd, 0)

  IF @rr.btnColor = @rr.PalClr(5) + &H080808 THEN 'if original color, change accordingly
     @rr.btnColor = GetSysColor(%COLOR_BTNFACE) + &H080808
  ELSEIF @rr.btnColor = @rr.PalClr(5) THEN         'common case - standard color..
     @rr.btnColor = GetSysColor(%COLOR_BTNFACE)
  END IF

  @rr.PalClr(1) = @rr.btnColor
  @rr.PalClr(2) = GetSysColor(%COLOR_BTNFACE) - &H101010
  @rr.PalClr(3) = GetSysColor(%COLOR_INFOBK)    '&HE1FFFF
  @rr.PalClr(4) = GetSysColor(%COLOR_SCROLLBAR) '&HC0C0C0
  @rr.PalClr(5) = GetSysColor(%COLOR_BTNFACE)   'for reference only

  hDC = GetDc(hWnd)                             'need own palette if in 256 color mode
  IF GetDeviceCaps(hDC, %NUMCOLORS) > -1 AND _
  (GetDeviceCaps(hDC, %RASTERCAPS) AND %RC_PALETTE) = %RC_PALETTE THEN
     IF @rr.PalClr(0) THEN DeleteObject @rr.PalClr(0)
     @rr.PalClr(0) = MAKLNG(&H0300, 4)
     @rr.PalClr(0) = CreatePalette (BYVAL VARPTR(@rr.PalClr(0)))
     FOR ii = 1 TO 4
        @rr.PalClr(ii) = @rr.PalClr(ii) + &H02000000
     NEXT
     @rr.btnColor = @rr.PalClr(1)
  END IF

  ReleaseDc hWnd, hDC

  IF @rr.hBrush THEN DeleteObject @rr.hBrush
  @rr.hBrush = CreateSolidBrush(@rr.btnColor)

END SUB

'
' Custom PtInRegion function
'------------------------------------------------------------------------------
FUNCTION CursorPtInRegion(BYVAL hWnd AS DWORD) AS LONG
  LOCAL rc AS RECT, pt AS POINTAPI, rr AS RRBUTTONDATA PTR

  GetWindowRect hWnd, rc   'get control's pos on screen
  GetCursorPos pt          'get cursor pos on screen

  IF PtInRect(rc, pt.x, pt.y) THEN
     ScreenToClient hWnd, pt      'convert cursor pos to client coordinates
     rr = GetWindowLong(hWnd, 0)  'get control specific data

     FUNCTION = PtInRegion(@rr.hRgn, pt.x, pt.y) 'see if it's inside region and return result
  END IF

END FUNCTION

'
' Draw button
'------------------------------------------------------------------------------
SUB DrawRRButton(BYVAL hWnd AS DWORD, BYVAL memDC AS DWORD, rc AS RECT)
  LOCAL hPen AS DWORD, LT AS LONG, RB AS LONG, col AS LONG, rr AS RRBUTTONDATA PTR
  rr = GetWindowLong(hWnd, 0)  'get control specific data

  IF @rr.labelMode = 2 OR _  ' BUTTON UP
     (@rr.labelMode = 0 AND (@rr.dn = 1 OR (@rr.raised = %TRUE AND @rr.dn = 0))) THEN
     CALL SendMessage(@rr.hParent, %RBN_BTNUP, @rr.hWnd, @rr.id)
     LT = 0 : RB = 1 : col = GetSysColor(%COLOR_BTNHILIGHT) : GOSUB DrawProcedure
     LT = 1 : RB = 0 : col = GetSysColor(%COLOR_BTNSHADOW)  : GOSUB DrawProcedure
     LT = 1 : RB = 1 : col = @rr.btnColor : GOSUB DrawProcedure
     IF @rr.FocusRect AND GetFocus = @rr.hWnd THEN
        IF @rr.raised OR @rr.labelMode THEN
           LT = 2 : RB = 2 : col = @rr.PalClr(2) : GOSUB DrawProcedure
        END IF
     END IF

  ELSEIF @rr.labelMode = 3 OR (@rr.labelMode = 0 AND @rr.dn = 2) THEN ' BUTTON DOWN
     CALL SendMessage(@rr.hParent, %RBN_BTNDN, @rr.hWnd, @rr.id)
     LT = 1 : RB = 0 : col = GetSysColor(%COLOR_BTNHILIGHT) : GOSUB DrawProcedure
     LT = 0 : RB = 1 : col = GetSysColor(%COLOR_BTNSHADOW) : GOSUB DrawProcedure
     LT = 1 : RB = 1 : col = @rr.btnColor : GOSUB DrawProcedure
     IF @rr.FocusRect AND @rr.labelMode AND GetFocus = @rr.hWnd THEN
        LT = 2 : RB = 2 : col = GetSysColor(%COLOR_INFOBK) : GOSUB DrawProcedure '@rr.PalClr(2)
     END IF

  ELSE 'FLAT MODE
     IF @rr.raised = 0 THEN CALL SendMessage(@rr.hParent, %RBN_BTNFLAT, @rr.hWnd, @rr.id)
     IF @rr.labelMode = 0 THEN
        LT = 0 : RB = 1
     ELSE
        LT = 0 : RB = 0
     END IF
     col = @rr.btnColor : GOSUB DrawProcedure
     IF @rr.FocusRect AND @rr.raised = 0 AND GetFocus = @rr.hWnd THEN
        LT = 0 : RB = 1
        IF @rr.labelMode = 0 THEN
           col = GetSysColor(%COLOR_INFOBK)
        ELSE
           col = @rr.PalClr(2)
        END IF
        GOSUB DrawProcedure
     END IF
  END IF

EXIT SUB

DrawProcedure:
  hPen = CreatePen(%PS_SOLID, 1, col)
  hPen = SelectObject(memDC, hPen)
  RoundRect memDC, LT, LT, rc.nright - RB, rc.nbottom - RB, @rr.corner, @rr.corner
  DeleteObject SelectObject(memDC, hPen)
  RETURN

END SUB
